package Anarres::Mud::Driver::Program::Type;

use strict;
use vars qw(@ISA @EXPORT_OK %EXPORT_TAGS);
use Exporter;
use Carp qw(:DEFAULT cluck);
use Data::Dumper;

BEGIN {
	no strict qw(refs);

	my %types = (
		T_VOID		=> "V",
		T_NIL		=> "N",
		T_MIXED		=> "_",
		T_BOOL		=> "b",
		T_CLOSURE	=> "f",
		T_INTEGER	=> "i",
		T_MAPPING	=> "m",
		T_OBJECT	=> "o",
		T_STRING	=> "s",
		T_ARRAY		=> "P_",
			);
	my $package = __PACKAGE__;

	foreach (keys %types) {
		# Why doesn't this work to let me reference the variable
		# as \$$_ in the eval below?
		# *{"${package}::$_"} = \${"${package}::$_"};		# "use vars"


		my $tpack = lc $_;
		$tpack =~ s/^T_//i;
		$tpack = ucfirst $tpack;

		eval qq{
			 *${package}::$_ = sub () { \$${package}::$_; };
			\@${package}::${tpack}::ISA = qw($package);
			\$${package}::$_ =
				bless \\\${types{$_}}, "${package}::${tpack}";
		}; die $@ if $@;

			# *${package}::$_ = \\\$${package}::$_;
			# my \$self = "$types{$_}";
			# return bless \\\$self, "$package";

		# print "$_: ${$_} = $${$_}\n";
	}

	# print Dumper(\%{"${package}::"});

	@ISA = qw(Exporter);
	@EXPORT_OK = keys %types;
	%EXPORT_TAGS = ( types => \@EXPORT_OK );
}

sub struct {
	my ($class, @fields) = @_;			# @fields are Types
	my $self = "S{" . join(",", map { $$_ } @fields) . "}";
	return bless \$self, $class;
}

sub equals {
	my ($self, $t) = @_;
	return $$self eq $$t;
}

sub promote {
	my ($self, $node, $type) = @_;

	q[
		int => bool (defined and nonzero)
		(all others just definedness)
		array => bool
		mapping => bool
		object => bool
		string => bool (definedness)
		closure => bool
		nil => bool (false)

		int => string (for addition, hash deref, assignment)
		object => string (for hash deref)
	] if 0;

	return $node if $$self eq $$type;

	my $name = ref($node);
	$name =~ s/^.*:://;
	print "Promote $name from $$self to $$type\n";

	return $node;
}

sub meet {
	my ($self, $other) = @_;
	return $self unless $other;

#	my $re = join("(?:", split(//, $$other));
#	$re = "((?:" . $re . (")?" x length $$other) . ")";
#
#	# print "$re\n";
#	$$self =~ /$re/;
#	my $match = $1;

	# This is cool.
	(my $t = $$self ^ $$other) =~ /^\0*/g;
	my $match = substr($$self, 0, pos($t));

	$match .= "_" if length $match < length $$other;

	print "Match $$self =~ $$other: $match\n";

	return bless \$match, ref($self);

	return $self;
}

sub classp { return ${$_[0]} =~ /^S/; }
sub arrayp { return ${$_[0]} =~ /^P/; }

sub pointer {
	my ($self, $stars) = @_;	# $type is a Type
	$stars or $stars = 1;
	my $type = ("P" x $stars) . $$self;
	return bless \$type, ref($self);
}

sub deref {
	my ($self) = @_;
	my $class = __PACKAGE__;

	if ($$self eq '_') {
		my $out = "_";
		return bless \$out, $class;
	}
	elsif ($$self eq 's') {
		return T_INTEGER;
	}
	elsif ($$self =~ /^P/) {
		my $out = substr($$self, 1);
		# print "Array deref $$self to $out : $class\n";
		return bless \$out, $class;
	}
	confess "Can't deref type $$self";
}

sub dump { return ${$_[0]}; }

1;
